//
//  ViewControllerMain.mm
//  StarGenerator
//
//  Created by song on 2016-02-25.
//  Copyright © 2016 song. All rights reserved.
//

#import "ViewControllerMain.h"
#import <vector>

@implementation ViewControllerMain

// constants //////////////////////////////////////////////////////////////////
const int STAR_POINTS = 5;      // default point
const int STAR_RADIUS = 10;     // default out rdius

///////////////////////////////////////////////////////////////////////////////
@synthesize modelGL;
@synthesize viewGL;
@synthesize points;
@synthesize checkFill;
@synthesize checkGrid;
@synthesize checkEdge;
@synthesize checkPoint;
@synthesize buttonCamera;
@synthesize buttonAbout;
@synthesize sliderPoints;
@synthesize sliderOutRadius;
@synthesize sliderInRadius;
@synthesize labelPoints;
@synthesize labelOutRadius;
@synthesize labelInRadius;
@synthesize tablePoints;



- (void)viewDidLoad
{
    [super viewDidLoad];

    // tell view component which model component is
    [viewGL setModel:&modelGL];
    
    // OpenGL RC (Rendering Context) should be ready at this point
    // init OpenGL states here
    modelGL.init();
    modelGL.setStar(STAR_POINTS, STAR_RADIUS);
    modelGL.setStarSelectedPoint(-1);
    printf("Initialized OpenGL states.\n");

    // display tableview
    points = [[NSMutableArray alloc] init];
    [tablePoints setDataSource:self];
    [self setTablePoints];

    // get size of opengl window, and pass it to model component
    NSRect rect = [viewGL bounds];
    modelGL.setWindowSize((int)rect.size.width, (int)rect.size.height);

    // init inner radius slider
    sliderInRadius.floatValue = modelGL.getStarInnerRadius();
    labelInRadius.stringValue = [NSString stringWithFormat:@"%.1f", sliderInRadius.floatValue];

    // draw first time
    [viewGL setNeedsDisplay:YES];
}



- (void)setRepresentedObject:(id)representedObject
{
    [super setRepresentedObject:representedObject];

    // Update the view, if already loaded.
}



///////////////////////////////////////////////////////////////////////////////
- (IBAction)toggleFill:(id)sender
{
    if(checkFill.state == NSOnState)
        modelGL.enableFill();
    else
        modelGL.disableFill();
    [viewGL setNeedsDisplay:YES];
}

- (IBAction)toggleGrid:(id)sender
{
    if(checkGrid.state == NSOnState)
        modelGL.enableGrid();
    else
        modelGL.disableGrid();
    [viewGL setNeedsDisplay:YES];
}

- (IBAction)toggleEdge:(id)sender
{
    if(checkEdge.state == NSOnState)
        modelGL.enableEdge();
    else
        modelGL.disableEdge();
    [viewGL setNeedsDisplay:YES];
}

- (IBAction)togglePoints:(id)sender
{
    if(checkPoint.state == NSOnState)
        modelGL.enablePoint();
    else
        modelGL.disablePoint();
    [viewGL setNeedsDisplay:YES];
}



- (IBAction)buttonCameraClicked:(id)sender
{
    modelGL.resetCamera();
    [viewGL setNeedsDisplay:YES];
}



- (IBAction)tablePointsClicked:(id)sender
{
    int index = (int)[tablePoints clickedRow];
    modelGL.setStarSelectedPoint(index);
    [viewGL setNeedsDisplay:YES];
     //printf("%d\n", index);
}



///////////////////////////////////////////////////////////////////////////////
- (IBAction)sliderPointsChanged:(id)sender
{
    labelPoints.stringValue = [NSString stringWithFormat:@"%d", sliderPoints.intValue];
    modelGL.setStarPointCount(sliderPoints.intValue);
    
    sliderInRadius.floatValue = modelGL.getStarInnerRadius();
    labelInRadius.stringValue = [NSString stringWithFormat:@"%.1f", sliderInRadius.floatValue];
    
    [self setTablePoints];
    [viewGL setNeedsDisplay:YES];
}

- (IBAction)sliderOutRadiusChanged:(id)sender
{
    labelOutRadius.stringValue = [NSString stringWithFormat:@"%.1f", sliderOutRadius.floatValue];
    modelGL.setStarOuterRadius(sliderOutRadius.floatValue);
    
    sliderInRadius.floatValue = modelGL.getStarInnerRadius();
    labelInRadius.stringValue = [NSString stringWithFormat:@"%.1f", sliderInRadius.floatValue];
    
    [self setTablePoints];
    [viewGL setNeedsDisplay:YES];
}

- (IBAction)sliderInRadiusChanged:(id)sender
{
    labelInRadius.stringValue = [NSString stringWithFormat:@"%.1f", sliderInRadius.floatValue];
    modelGL.setStarInnerRadius(sliderInRadius.floatValue);
    [self setTablePoints];
    [viewGL setNeedsDisplay:YES];
}



///////////////////////////////////////////////////////////////////////////////
- (void)prepareForSegue:(NSStoryboardSegue*)segue sender:(id)sender
{
    if([[segue identifier] isEqualToString:@"segueAbout"])
    {
        // set delegate before open about sheet
        ViewControllerAbout* ctrl = [segue destinationController];
        ctrl.delegate = self;
        ctrl.starMaxPoints = sliderPoints.maxValue;     // pass current value
        ctrl.starMaxRadius = sliderOutRadius.maxValue;
    }
}



///////////////////////////////////////////////////////////////////////////////
- (void)setTablePoints
{
    [points removeAllObjects];
    
    std::vector<Vector2> p = modelGL.getStarPoints();
    int count = (int)p.size();
    for(int i = 0; i < count; ++i)
    {
        NSString* row = [NSString stringWithFormat:@"%d: (%.1f, %.1f)", i+1, p[i].x, p[i].y];
        [points addObject:row];
    }
    [tablePoints reloadData];

    // reset prev selection
    if(count <= (int)[tablePoints clickedRow])
    {
        modelGL.setStarSelectedPoint(-1);
        [tablePoints deselectAll:nil];
    }
}



///////////////////////////////////////////////////////////////////////////////
// delegate for NSTableViewDataSource
- (NSInteger)numberOfRowsInTableView:(NSTableView*)tableView
{
    return [points count];
}
     
- (id)tableView:(NSTableView*)tableView objectValueForTableColumn:(NSTableColumn*)column row:(NSInteger)row
{
    return [points objectAtIndex:row];
}



///////////////////////////////////////////////////////////////////////////////
// delegate for AboutDelegate
- (void)aboutClosed:(ViewControllerAbout*)sender
{
    int currentPointCount = sliderPoints.intValue;
    float currentRadius = sliderOutRadius.floatValue;
    if(currentPointCount > sender.starMaxPoints)
        modelGL.setStarPointCount(sender.starMaxPoints);
    if(currentRadius > sender.starMaxRadius)
        modelGL.setStarOuterRadius(sender.starMaxRadius);

    sliderPoints.maxValue = sender.starMaxPoints;
    sliderOutRadius.maxValue = sender.starMaxRadius;
    sliderInRadius.maxValue = sender.starMaxRadius;
    
    labelPoints.stringValue = [NSString stringWithFormat:@"%d", sliderPoints.intValue];
    labelOutRadius.stringValue = [NSString stringWithFormat:@"%.1f", sliderOutRadius.floatValue];
    labelInRadius.stringValue = [NSString stringWithFormat:@"%.1f", sliderInRadius.floatValue];

    // resize grid
    modelGL.setGridSize(sender.starMaxRadius);
    [self setTablePoints];
    [viewGL setNeedsDisplay:YES];
}


@end
